library line_break;
import 'dart:math';
import 'package:line_break/properties/line_break_property.dart';
import 'package:line_break/properties/line_break_properties.dart';

enum LineBreakAction {
  direct,
  indirect,
  combiningIndirect,
  prohibited,
  combiningProhibited,
}

enum LineBreakResult {
  //必须换行
  must,
  //允许换行
  allowed,
  //禁止换行
  prohibited,
  //默认
  none,
}

class LineBreak {

  List pairTable = List<List<LineBreakAction>>(33);
  DefaultProperties defaultProperties = DefaultProperties();
  ChineseProperties chineseProperties = ChineseProperties();

  LineBreak() {
    _initPairTable();
  }

  HashTableProperties getProperties([String lang]) {
    switch (lang) {
      case "zh": {
        return chineseProperties;
      }
      default: {
        return defaultProperties;
      }
    }
  }

  LineBreakClass getClassForValue(int value, String lang) {
    LineBreakClass lineBreakClass = getProperties(lang).getClassForValue(value);
    if (lineBreakClass == LineBreakClass.XX) {
      return defaultProperties.getClassForValue(value);
    }
    return lineBreakClass;
  }

  LineBreakClass resolveLineBreakClass(LineBreakClass lineBreakClass, [String lang]) {
    LineBreakClass result = LineBreakClass.Undefined;
    switch (lineBreakClass) {
      case LineBreakClass.AI:
        result = LineBreakClass.AL;
        if (lang == "zh" || lang == "ko" || lang == "ja") {
          result = LineBreakClass.ID;
        }
        break;
      case LineBreakClass.SA:
      case LineBreakClass.SG:
      case LineBreakClass.XX:
        result = LineBreakClass.AL;
        break;
      case LineBreakClass.CJ:
        if (lang.endsWith('-strict')){
          result = LineBreakClass.NS;
        } else {
          result = LineBreakClass.ID;
        }
        break;
      default:
        result = lineBreakClass;
        break;
    }
    return result;
  }

  List<LineBreakResult> findLineBreaks(String text, [String lang]) {

    LineBreakClass firstClass = getClassForValue(text.codeUnitAt(0), lang);
    bool fLb8aZwj = firstClass == LineBreakClass.ZWJ; /**< Flag for ZWJ (LB8a) */
    bool fLb10LeadSpace = firstClass == LineBreakClass.SP; /**< Flag for leading space (LB10) */
    bool fLb21aHebrew = false;              /**< Flag for Hebrew letters (LB21a) */
    int cLb30aRI = 0;                   /**< Count of RI characters (LB30a) */

    LineBreakClass currentClass = resolveLineBreakClass(firstClass, lang);
    LineBreakClass prevClass = LineBreakClass.Undefined;
    LineBreakClass nextClass = LineBreakClass.Undefined;
    int length = text.codeUnits.length;

    List<LineBreakResult> results = List.filled(length, LineBreakResult.none, growable: true);


    for (int i = 1; i < length; i++) {

      switch (currentClass) {
        case LineBreakClass.NL:
        case LineBreakClass.LF:
          currentClass = LineBreakClass.BK;
          break;
        case LineBreakClass.SP:
          currentClass = LineBreakClass.WJ;
          break;
        default: break;
      }

      prevClass = nextClass;
      nextClass = getClassForValue(text.codeUnitAt(i), lang);

      if (currentClass == LineBreakClass.BK ||
          (currentClass == LineBreakClass.CR && nextClass != LineBreakClass.LF)) {
        results.replaceRange(i, i + 1, [LineBreakResult.must]);
        currentClass = getClassForValue(text.codeUnitAt(i), lang);
        continue;
      }

      LineBreakResult result = LineBreakResult.prohibited;
      bool isContinue = false;

      switch(nextClass) {
        case LineBreakClass.SP:  // Rule LB7
          result = LineBreakResult.prohibited;
          isContinue = true;
          break;
        case LineBreakClass.BK: // Rule LB6
        case LineBreakClass.NL:
        case LineBreakClass.LF:
          result = LineBreakResult.prohibited;
          currentClass = LineBreakClass.BK;
          isContinue = true;
          break;
        case LineBreakClass.CR: // Rule LB6
          result = LineBreakResult.prohibited;
          currentClass = LineBreakClass.CR;
          isContinue = true;
          break;
        default: break;
      }


      if (isContinue == false) {
        nextClass = resolveLineBreakClass(nextClass, lang);
        LineBreakAction action = pairTable[currentClass.index][nextClass.index];

        switch (action) {
          case LineBreakAction.direct:
            result = LineBreakResult.allowed;
            break;
          case LineBreakAction.indirect:
          case LineBreakAction.combiningIndirect:
            result = (prevClass == LineBreakClass.SP) ? LineBreakResult.allowed : LineBreakResult.prohibited;
            break;
          case LineBreakAction.combiningProhibited:
            if (prevClass == LineBreakClass.SP) {
              result = LineBreakResult.prohibited;
            }
            break;
          case LineBreakAction.prohibited:
            result = LineBreakResult.prohibited;
            break;
        }
      }

      if (fLb8aZwj) {
        result = LineBreakResult.prohibited;
      }

      if (fLb21aHebrew &&
          (currentClass == LineBreakClass.HY || currentClass == LineBreakClass.BA)) {
        result = LineBreakResult.prohibited;
        fLb21aHebrew = false;
      } else {
        fLb21aHebrew = currentClass == LineBreakClass.HL;
      }

      if (currentClass == LineBreakClass.RI) {
        cLb30aRI++;
        if (cLb30aRI == 2 && nextClass == LineBreakClass.RI) {
          result = LineBreakResult.allowed;
          cLb30aRI = 0;
        }
      } else {
        cLb30aRI = 0;
      }

      currentClass = nextClass;

      fLb8aZwj = nextClass == LineBreakClass.ZWJ;
      if (fLb10LeadSpace == true) {
        if (nextClass == LineBreakClass.CM || nextClass == LineBreakClass.ZWJ) {
          result = LineBreakResult.allowed;
        }
        fLb10LeadSpace = false;
      }

      results.replaceRange(i - 1, i, [result]);
    }

    results.replaceRange(length - 1, length, [LineBreakResult.must]);
    return results;
  }

  /*
*  convert unicode data to utf16 data
* */
  List<int> unicode2utf16(List<int> wordArray) {
    List<int> list = new List()..length = wordArray.length * 4;
    int pos = 0;

    for (int i = 0, c = wordArray.length; i < c; ++i) {
      int word = wordArray[i];
      if (word > 0xffff) {
        var c  = word - 0x10000;
        var sh = (c >> 10) + 0xD800;
        var sl = (c & 0xFFF) + 0xDC00;
        list[pos++] = sh >> 8;
        list[pos++] = sh & 0xFF;
        list[pos++] = sl >> 8;
        list[pos++] = sh & 0xFF;
      } else {
        list[pos++] = word >> 8;
        list[pos++] = word & 0xFF;
      }
    }

    list.length = pos;
    return list;
  }

  void _initPairTable() {

    LineBreakAction direct = LineBreakAction.direct;
    LineBreakAction indirect = LineBreakAction.indirect;
    LineBreakAction combiningIndirect = LineBreakAction.combiningIndirect;
    LineBreakAction prohibited = LineBreakAction.prohibited;
    LineBreakAction combiningProhibited = LineBreakAction.combiningProhibited;

    int OP = LineBreakClass.OP.index - 1;
    int CL = LineBreakClass.CL.index - 1;
    int CP = LineBreakClass.CP.index - 1;
    int QU = LineBreakClass.QU.index - 1;
    int GL = LineBreakClass.GL.index - 1;
    int NS = LineBreakClass.NS.index - 1;
    int EX = LineBreakClass.EX.index - 1;
    int SY = LineBreakClass.SY.index - 1;
    int IS = LineBreakClass.IS.index - 1;
    int PR = LineBreakClass.PR.index - 1;
    int PO = LineBreakClass.PO.index - 1;
    int NU = LineBreakClass.NU.index - 1;
    int AL = LineBreakClass.AL.index - 1;
    int HL = LineBreakClass.HL.index - 1;
    int ID = LineBreakClass.ID.index - 1;
    int IN = LineBreakClass.IN.index - 1;
    int HY = LineBreakClass.HY.index - 1;
    int BA = LineBreakClass.BA.index - 1;
    int BB = LineBreakClass.BB.index - 1;
    int B2 = LineBreakClass.B2.index - 1;
    int ZW = LineBreakClass.ZW.index - 1;
    int CM = LineBreakClass.CM.index - 1;
    int WJ = LineBreakClass.WJ.index - 1;
    int H2 = LineBreakClass.H2.index - 1;
    int H3 = LineBreakClass.H3.index - 1;
    int JL = LineBreakClass.JL.index - 1;
    int JV = LineBreakClass.JV.index - 1;
    int JT = LineBreakClass.JT.index - 1;
    int RI = LineBreakClass.RI.index - 1;
    int EB = LineBreakClass.EB.index - 1;
    int EM = LineBreakClass.EM.index - 1;
    int ZWJ = LineBreakClass.ZWJ.index - 1;
    int CB = LineBreakClass.CB.index - 1;

    pairTable[OP] = [   /* OP */
      prohibited, prohibited, prohibited, prohibited, prohibited, prohibited, prohibited,
      prohibited, prohibited, prohibited, prohibited, prohibited, prohibited, prohibited,
      prohibited, prohibited, prohibited, prohibited, prohibited, prohibited, prohibited,
      combiningProhibited, prohibited, prohibited, prohibited, prohibited, prohibited, prohibited,
      prohibited, prohibited, prohibited, prohibited, prohibited];

    pairTable[CL] = [   /* CL */
      direct, prohibited, prohibited, indirect, indirect, prohibited, prohibited,
      prohibited, prohibited, indirect, indirect, direct, direct, direct,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[CP] = [   /* CP */
      direct, prohibited, prohibited, indirect, indirect, prohibited, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[QU] = [   /* QU */
      prohibited, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect, indirect, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect];

    pairTable[GL] = [   /* GL */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect, indirect, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect];

    pairTable[NS] = [   /* NS */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[EX] = [   /* EX */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[SY] = [   /* SY */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, indirect, direct, indirect,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[IS] = [   /* IS */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, indirect, indirect, indirect,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[PR] = [   /* PR */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, indirect, indirect, indirect,
      indirect, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct];

    pairTable[PO] = [   /* PO */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, indirect, indirect, indirect,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[NU] = [   /* NU */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[AL] = [   /* AL */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[HL] = [   /* HL */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[ID] = [   /* ID */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[IN] = [   /* IN */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[HY] = [   /* HY */
      direct, prohibited, prohibited, indirect, direct, indirect, prohibited,
      prohibited, prohibited, direct, direct, indirect, direct, direct,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[BA] = [   /* BA */
      direct, prohibited, prohibited, indirect, direct, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[BB] = [   /* BB */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect, indirect, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, direct];

    pairTable[B2] = [   /* B2 */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, direct, indirect, indirect, direct, prohibited, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[ZW] = [   /* ZW */
      direct, direct, direct, direct, direct, direct, direct,
      direct, direct, direct, direct, direct, direct, direct,
      direct, direct, direct, direct, direct, direct, prohibited,
      direct, direct, direct, direct, direct, direct, direct,
      direct, direct, direct, direct, direct];

    pairTable[CM] = [   /* CM */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[WJ] = [   /* WJ */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect, indirect, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, indirect,
      indirect, indirect, indirect, indirect, indirect];

    pairTable[H2] = [   /* H2 */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, indirect, indirect,
      direct, direct, direct, indirect, direct];

    pairTable[H3] = [   /* H3 */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, indirect,
      direct, direct, direct, indirect, direct];

    pairTable[JL] = [   /* JL */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, indirect, indirect, indirect, indirect, direct,
      direct, direct, direct, indirect, direct];

    pairTable[JV] = [   /* JV */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, indirect, indirect,
      direct, direct, direct, indirect, direct];

    pairTable[JT] = [   /* JT */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, indirect,
      direct, direct, direct, indirect, direct];

    pairTable[RI] = [   /* RI */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, direct, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      indirect, direct, direct, indirect, direct];

    pairTable[EB] = [   /* EB */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, indirect, indirect, direct];

    pairTable[EM] = [   /* EM */
      direct, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, direct, indirect, direct, direct, direct,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[ZWJ] = [   /* ZWJ */
      indirect, prohibited, prohibited, indirect, indirect, indirect, prohibited,
      prohibited, prohibited, indirect, indirect, indirect, indirect, indirect,
      direct, indirect, indirect, indirect, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];

    pairTable[CB] = [   /* CB */
      direct, prohibited, prohibited, indirect, indirect, direct, prohibited,
      prohibited, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, direct, direct, direct, prohibited,
      combiningIndirect, prohibited, direct, direct, direct, direct, direct,
      direct, direct, direct, indirect, direct];
  }
}


void main() {
  LineBreak lineBreak = LineBreak();
  List<LineBreakResult> result = lineBreak.findLineBreaks(r"这个语言\n还是比较奇怪的。 之前根本没听过（可能是我孤陋寡闻了），据说有人问 flutter 开发者，为什么使用 dart 语言。开发者回答说：dart 开发组就在我们隔壁，所以我们决定使用 dart", "zh");
  print(result);
}